from cytomat.serial_port import SerialPort
from cytomat.status import OverviewStatus


class PlateHandler:
    __serial_port: SerialPort

    def __init__(self, serial_port: SerialPort) -> None:
        self.__serial_port = serial_port

    def initialize(self) -> OverviewStatus:
        """(Re-) initialize the plate handler"""
        return self.__serial_port.issue_action_command("ll:in")

    def move_plate_from_transfer_station_to_slot(self, slot: int) -> OverviewStatus:
        """
        Move a plate from the transfer station to the given slot

        Parameters
        ----------
        slot
            The target slot
        """
        return self.__serial_port.issue_action_command(f"mv:ts {slot:03}")

    def move_plate_from_slot_to_transfer_station(self, slot: int) -> OverviewStatus:
        """
        Move a plate from the given slot to the transfer station

        Parameters
        ----------
        slot
            The slot where the plate is located
        """
        return self.__serial_port.issue_action_command(f"mv:st {slot:03}")

    def move_plate_from_transfer_station_to_handler(self) -> OverviewStatus:
        """
        Move a plate from the transfer station to the plate handler shovel
        """
        return self.__serial_port.issue_action_command("mv:tw")

    def move_plate_from_handler_to_transfer_station(self) -> OverviewStatus:
        """
        Move a plate from the plate handler shovel to the transfer station
        """
        return self.__serial_port.issue_action_command("mv:wt")

    def move_plate_from_exposed_position_to_inside(self) -> OverviewStatus:
        """
        Move a plate from the exposed position (above the transfer station) to the neutral position inside the device
        """
        return self.__serial_port.issue_action_command("mv:hw")

    def move_plate_from_inside_to_exposed_position(self) -> OverviewStatus:
        """
        Move a plate from the neutral position inside the device to the exposed position (above the transfer station)
        """
        return self.__serial_port.issue_action_command("mv:wh")

    def move_plate_from_handler_to_slot(self, slot: int) -> OverviewStatus:
        """
        Move a plate from the plate handler shovel to the given slot

        Parameters
        ----------
        slot
            The target slot
        """
        return self.__serial_port.issue_action_command(f"mv:ws {slot:03}")

    def move_plate_from_slot_to_handler(self, slot: int) -> OverviewStatus:
        """
        Move a plate from the given slot to the plate handler shovel

        Parameters
        ----------
        slot
            The slot where the plate is located
        """
        return self.__serial_port.issue_action_command(f"mv:sw {slot:03}")

    def move_plate_from_exposed_position_to_slot(self, slot: int) -> OverviewStatus:
        """
        Move a plate from the exposed position (above the transfer station) to the given slot

        Parameters
        ----------
        slot
            The target slot
        """
        return self.__serial_port.issue_action_command(f"mv:hs {slot:03}")

    def move_plate_from_slot_to_exposed_position(self, slot: int) -> OverviewStatus:
        """
        Move a plate from the given slot to the exposed position (above the transfer station)

        Parameters
        ----------
        slot
            The slot where the plate is located
        """
        return self.__serial_port.issue_action_command(f"mv:sh {slot:03}")

    def retract_shovel(self) -> OverviewStatus:
        """
        Retract the plate handler shovel
        """
        return self.__serial_port.issue_action_command("ll:sp 001")

    def extend_shovel(self) -> OverviewStatus:
        """Extend the plate handler shovel"""
        return self.__serial_port.issue_action_command("ll:sp 002")

    def close_transfer_door(self) -> OverviewStatus:
        """Close the transfer door"""
        return self.__serial_port.issue_action_command("ll:gp 001")

    def open_transfer_door(self) -> OverviewStatus:
        """Open the transfer door"""
        return self.__serial_port.issue_action_command("ll:gp 002")

    def reset_handler_position(self) -> OverviewStatus:
        """Reset the handler to the neutral position"""
        return self.__serial_port.issue_action_command("ll:wp")

    def move_handler_below_slot_height(self, slot: int) -> OverviewStatus:
        """
        Move the handler below the given slot (only changes height, not rotation)

        Parameters
        ----------
        slot
            The target slot
        """
        return self.__serial_port.issue_action_command(f"ll:h- {slot:03}")

    def move_handler_above_slot_height(self, slot: int) -> OverviewStatus:
        """
        Move the handler above the given slot (only changes height, not rotation)

        Parameters
        ----------
        slot
            The target slot
        """
        return self.__serial_port.issue_action_command(f"ll:h+ {slot:03}")

    def rotate_handler_to_slot(self, slot: int) -> OverviewStatus:
        """
        Rotate the handler to the given slot (only changes rotation, not height)

        Parameters
        ----------
        slot
            The target slot
        """
        return self.__serial_port.issue_action_command(f"ll:dp {slot:03}")

    def rotate_handler_to_transfer_station(self) -> OverviewStatus:
        """
        Rotate the handler to the transfer station (only changes rotation, not height)
        """
        return self.rotate_handler_to_slot(0)

    def move_x_to_slot(self, slot: int) -> OverviewStatus:
        """
        Move along to the given slot (only moves along the x axis)

        Parameters
        ----------
        slot
            The target slot
        """
        return self.__serial_port.issue_action_command(f"ll: {slot:03}")



from __future__ import annotations


class UnexpectedResponse(Exception):
    pass


class InvalidCommand(Exception):
    pass


class SerialCommunicationError(Exception):
    """Errors corresponding to device responses 'er XX'"""

    @staticmethod
    def from_error_code(code: int) -> SerialCommunicationError:
        error_code_mapping = {
            0x01: DeviceBusy,
            0x02: CommandUnknown,
            0x03: InvalidCommandTelegramStructure,
            0x04: WrongTelegramParameters,
            0x05: UnknownSlotNumber,
            0x11: WrongHandlerPosition,
            0x12: ShovelNotExtended,
            0x21: HandlerOccupied,
            0x22: HandlerEmpty,
            0x31: TransferStationEmpty,
            0x32: TransferStationOccupied,
            0x33: TransferStationNotInPosition,
            0x41: TransferDoorNotConfigured,
            0x42: TransferDoorNotOpen,
            0x51: InternalMemoryAccessError,
            0x52: UnauthorizedAccess,
            0x61: ShakerNotActive,
        }
        try:
            return error_code_mapping[code]()
        except KeyError:
            return UnknownErrorCode(f"Unknown hex error code: {code:x}")


class UnknownErrorCode(SerialCommunicationError):
    pass


class DeviceBusy(SerialCommunicationError):
    pass


class CommandUnknown(SerialCommunicationError):
    pass


class InvalidCommandTelegramStructure(SerialCommunicationError):
    pass


class WrongTelegramParameters(SerialCommunicationError):
    pass


class UnknownSlotNumber(SerialCommunicationError):
    pass


class WrongHandlerPosition(SerialCommunicationError):
    pass


class ShovelNotExtended(SerialCommunicationError):
    pass


class HandlerOccupied(SerialCommunicationError):
    pass


class HandlerEmpty(SerialCommunicationError):
    pass


class TransferStationEmpty(SerialCommunicationError):
    pass


class TransferStationNotInPosition(SerialCommunicationError):
    pass


class TransferDoorNotConfigured(SerialCommunicationError):
    pass


class TransferDoorNotOpen(SerialCommunicationError):
    pass


class TransferStationOccupied(SerialCommunicationError):
    pass


class InternalMemoryAccessError(SerialCommunicationError):
    pass


class UnauthorizedAccess(SerialCommunicationError):
    pass


class InvalidResponseTelegramStructure(SerialCommunicationError):
    pass


class Timeout(SerialCommunicationError):
    pass


class ShakerNotActive(SerialCommunicationError):
    pass



from cytomat.serial_port import SerialPort
from cytomat.status import OverviewStatus


class MaintenanceController:
    __serial_port: SerialPort

    def __init__(self, serial_port: SerialPort) -> None:
        self.__serial_port = serial_port

    def reset_error_status(self) -> OverviewStatus:
        """Reset the error status"""
        return self.__serial_port.issue_action_command("rs:be")

    def restart_device(self) -> OverviewStatus:
        """Restart the device"""
        return self.__serial_port.issue_action_command("se:ns")

    def set_pitch(self, stacker: int, pitch: int) -> OverviewStatus:
        """
        Set the pitch of the given stacker

        Parameters
        ----------
        stacker
            The stacker index (1-based)
        pitch
            The pitch (in steps)
        """
        return self.__serial_port.issue_action_command(f"se:cs {stacker:03} {pitch:03}")

    def send_barcode_scanner_data_via_rs232(self) -> OverviewStatus:
        """Send barcode scanner data via the RS-232 interface"""
        return self.__serial_port.issue_action_command("se:c1")

    def send_temperate_co2_via_rs232(self) -> OverviewStatus:
        """Send barcode scanner data via the RS-232 interface"""
        return self.__serial_port.issue_action_command("se:c2")



from cytomat.serial_port import SerialPort
from cytomat.status import OverviewStatus


class ShakerController:
    __serial_port: SerialPort

    def __init__(self, serial_port: SerialPort) -> None:
        self.__serial_port = serial_port

    def get_shaker_frequency(self, shaker: int) -> int:
        """
        Get the frequency of the given shaker

        Parameters
        ----------
        shaker
            The shaker number (1-based)

        Returns
        -------
        The shaker frequency
        """
        self.__check_shaker_id(shaker)
        return int(self.__serial_port.issue_status_command(f"ch:pb {shaker + 19}"))

    def set_shaker_frequency(self, shaker: int, frequency: int) -> OverviewStatus:
        """
        Set the frequency of the given shaker

        Parameters
        ----------
        shaker
            The shaker number (1-based)
        frequency
            The target frequency
        """
        self.__check_shaker_id(shaker)
        return self.__serial_port.issue_action_command(f"se:pb {shaker + 19} {frequency:04}")

    def initialize_shakers(self) -> OverviewStatus:
        """Initialize the shakers"""
        return self.__serial_port.issue_action_command("ll:vi")

    def start_all_shakers(self) -> OverviewStatus:
        """Start all shakers"""
        return self.__serial_port.issue_action_command("ll:va")

    def start_shaker(self, shaker: int) -> OverviewStatus:
        """
        Start the given shaker

        Parameters
        ----------
        shaker
            The shaker number (1-based)
        """
        self.__check_shaker_id(shaker)
        return self.__serial_port.issue_action_command(f"ll:va {shaker:03}")

    def stop_all_shakers(self) -> OverviewStatus:
        """Stop all shakers"""
        return self.__serial_port.issue_action_command("ll:vd")

    def stop_shaker(self, shaker: int) -> OverviewStatus:
        """
        Stop the given shaker

        Parameters
        ----------
        shaker
            The shaker number (1-based)
        """
        self.__check_shaker_id(shaker)
        return self.__serial_port.issue_action_command(f"ll:vd {shaker:03}")

    @staticmethod
    def __check_shaker_id(shaker: int) -> None:
        if shaker not in (1, 2):
            raise ValueError("Invalid shaker, must be 1 or 2")



import sys

if sys.version_info[:2] < (3, 8):
    import importlib_metadata as metadata
else:
    from importlib import metadata

from .cytomat import Cytomat

__version__ = metadata.version("open-cytomat")

__all__ = ["__version__", "Cytomat"]



from cytomat.serial_port import SerialPort


class ClimateController:
    __serial_port: SerialPort

    def __init__(self, serial_port: SerialPort) -> None:
        self.__serial_port = serial_port

    @property
    def current_temperature(self) -> float:
        """The current temperature, in °C (read-only)"""
        return float(self.__serial_port.issue_status_command("ch:it").split()[1])

    @property
    def target_temperature(self) -> float:
        """The target temperature, in °C (can be set)"""
        return float(self.__serial_port.issue_status_command("ch:it").split()[0])

    @target_temperature.setter
    def target_temperature(self, target: float) -> None:
        if target > 100 or target < 0:
            raise ValueError("Temperature must be > 0 and < 100")
        self.__serial_port.issue_action_command(f"ll:it {target:04.1f}")

    @property
    def current_co2(self) -> float:
        """The current CO2 level (read-only)"""
        return float(self.__serial_port.issue_status_command("ch:ic").split()[1])

    @property
    def target_co2(self) -> float:
        """The target CO2 level (can be set)"""
        return float(self.__serial_port.issue_status_command("ch:ic").split()[0])

    @target_co2.setter
    def target_co2(self, target: float) -> None:
        if target > 100 or target < 0:
            raise ValueError("CO2 must be > 0 and < 100")
        self.__serial_port.issue_action_command(f"ll:ic {target:04.1f}")



import re
from threading import Lock

from serial import Serial

from cytomat.errors import InvalidCommand, SerialCommunicationError, UnexpectedResponse
from cytomat.status import OverviewStatus
from cytomat.utils import lock_threading_lock


class SerialPort:
    __serial_port: Serial
    __port_lock: Lock
    __timeout: float

    def __init__(self, port: str, *, timeout: float) -> None:
        """
        Wraps a plain serial connection to simplify communication with Cytomat devices

        Parameters
        ----------
        port : str
            The address of the port (e.g. "COM3" on Windows or "/dev/ttyACM0" on Unix)
        timeout: float, seconds
            A :class:`TimeoutError` will be raised of read or write operations took longer than this duration
        """
        self.__serial_port = Serial(port, timeout=timeout, write_timeout=timeout)
        self.__port_lock = Lock()
        self.__timeout = timeout

    def __communicate(self, command: str) -> str:
        """
        Send a command to the port and return the response string

        Parameters
        ----------
        command: str
            The command to send

        Returns
        -------
        str
            The response

        Raises
        ------
        TimeoutError
            If a read or write operation took too long
        SerialCommunicationError
            If there were unread bytes in the input buffer before sending the command
        """
        with lock_threading_lock(self.__port_lock, timeout=self.__timeout):
            if self.__serial_port.in_waiting:
                raise SerialCommunicationError("There were unread bytes in the input buffer")

            raw_command: bytes = command.encode("ascii") + b"\r"
            raw_response: bytes = b""
            self.__serial_port.write(raw_command)
            while not raw_response.endswith(b"\r"):
                char = self.__serial_port.read()
                if not char:
                    raise TimeoutError(rf"Did not receive a '\r'-terminated response after {self.__timeout} seconds")
                raw_response += char

        response: str = raw_response[:-1].decode("ascii")
        return response

    @staticmethod
    def __check_prefix_and_strip(response: str, expected_prefix: str) -> str:
        """
        Check if the response has the expected prefix, strip the prefix and whitespace

        Parameters
        ----------
        response: str
            A response as sent by the device
        expected_prefix: str
            The expected prefix

        Returns
        -------
        str
            The response without the prefix and surrounding whitespace

        Raises
        ------
        UnexpectedResponse
            If the prefix did not match
        """
        if response.startswith(expected_prefix):
            return response[len(expected_prefix) :].strip()
        raise UnexpectedResponse(f"Expected response prefix '{expected_prefix}', got response '{response}'")

    def issue_action_command(self, command: str) -> OverviewStatus:
        """
        Issue a command which results in 'ok XX' or 'er XX'.

        Parameters
        ----------
        command: str
            The command

        Returns
        -------
        OverviewStatus
            The current device status, as given by the response 'ok XX'

        Raises
        ------
        SerialCommunicationError
            If the response was 'er XX'
        UnexpectedResponse
            If the response did not match 'ok XX' or 'er XX'
        """
        response = self.__communicate(command)

        if response.startswith("ok"):
            return OverviewStatus.from_hex_string(self.__check_prefix_and_strip(response, "ok"))
        if response.startswith("er"):
            raise SerialCommunicationError.from_error_code(int(self.__check_prefix_and_strip(response, "er"), base=16))
        raise UnexpectedResponse(f"Expected response like 'ok XX' or 'er XX', got '{response}'")

    def issue_status_command(self, command: str) -> str:
        """
        Issue a status command in the form of 'ch:xx' or 'ch:xx ...', which results in responses like 'xx ...'.
        Returns the response without the prefix

        Parameters
        ----------
        command: str
            The command

        Returns
        -------
        str
            The response without the prefix

        Raises
        ------
        InvalidCommand
            If the command did not match 'ch:xx' or 'ch:xx ...'

        Examples
        --------
        >>> self.issue_action_command("ch:bs")
        '01'  # full response was 'bs 01'
        """
        if not re.fullmatch("ch:[a-z]{2}.*", command):
            raise InvalidCommand(f"Expected command like 'ch:xx' or 'ch:xx ...', got '{command}'")

        return self.__check_prefix_and_strip(self.__communicate(command), expected_prefix=command[3:5])



from cytomat.serial_port import SerialPort
from cytomat.status import OverviewStatus, SwapStationStatus


class SwapStation:
    __serial_port: SerialPort

    def __init__(self, serial_port: SerialPort) -> None:
        self.__serial_port = serial_port

    @property
    def status(self) -> SwapStationStatus:
        """The swap station status"""
        return SwapStationStatus.from_response_string(self.__serial_port.issue_status_command("ch:sw"))

    def rotate_to_position_1(self) -> OverviewStatus:
        """Rotate to position 1"""
        return self.__serial_port.issue_action_command("ll:tp 001")

    def rotate_to_position_2(self) -> OverviewStatus:
        """Rotate to position 2"""
        return self.__serial_port.issue_action_command("ll:tp 002")



import time
from datetime import datetime, timedelta

from cytomat.barcode_scanner import BarcodeScanner
from cytomat.climate_controller import ClimateController
from cytomat.maintenance_controller import MaintenanceController
from cytomat.plate_handler import PlateHandler
from cytomat.serial_port import SerialPort
from cytomat.shaker_controller import ShakerController
from cytomat.status import ActionStatus, ErrorStatus, OverviewStatus, WarningStatus
from cytomat.utils import enum_to_dict


class Cytomat:
    __serial_port: SerialPort

    plate_handler: PlateHandler
    """
    Exposes functionality related to plate handling
    """
    barcode_scanner: BarcodeScanner
    """
    Exposes functionality of the built-in barcode scanner
    """
    maintenance_controller: MaintenanceController
    """
    Exposes maintenance functionality
    """
    climate_controller: ClimateController
    """
    Exposes functionality related to temperature and CO2
    """
    shaker_controller: ShakerController
    """
    Exposes tower shaker functionality
    """

    def __init__(self, serial_port: str):
        self.__serial_port = SerialPort(serial_port, timeout=1)

        self.plate_handler = PlateHandler(self.__serial_port)
        self.barcode_scanner = BarcodeScanner(self.__serial_port)
        self.maintenance_controller = MaintenanceController(self.__serial_port)
        self.climate_controller = ClimateController(self.__serial_port)
        self.shaker_controller = ShakerController(self.__serial_port)

    @property
    def overview_status(self) -> OverviewStatus:
        """Status overview"""
        return OverviewStatus.from_hex_string(self.__serial_port.issue_status_command("ch:bs"))

    @property
    def action_status(self) -> ActionStatus:
        """Action status"""
        return ActionStatus.from_hex_string(self.__serial_port.issue_status_command("ch:ba"))

    @property
    def error_status(self) -> ErrorStatus:
        """Error status"""
        return enum_to_dict(ErrorStatus)[int(self.__serial_port.issue_status_command("ch:be"), base=16)]

    @property
    def warning_status(self) -> WarningStatus:
        """Warning status"""
        return enum_to_dict(WarningStatus)[int(self.__serial_port.issue_status_command("ch:bw"), base=16)]

    def wait_until_not_busy(self, timeout: float, poll_interval: float = 0.5) -> OverviewStatus:
        """
        Block the current thread until the device is not busy anymore.

        Parameters
        ----------
        timeout
            The timeout in seconds
        poll_interval
            The polling interval in seconds

        Raises
        ------
        TimeoutError
            If the device is still busy after the given timeout duration
        """
        status = self.overview_status
        end_time = datetime.now() + timedelta(seconds=timeout)
        while status.busy:
            if end_time < datetime.now():
                raise TimeoutError(f"Device still busy after {timeout} seconds")
            time.sleep(poll_interval)
            status = self.overview_status
        return status



from __future__ import annotations

from enum import IntEnum
from typing import NamedTuple

from cytomat.utils import enum_to_dict, int_to_bits


class OverviewStatus(NamedTuple):
    busy: bool
    ready: bool
    warning: bool
    error: bool
    shovel_occupied: bool
    transfer_door_open: bool
    device_door_open: bool
    transfer_station_occupied: bool

    @classmethod
    def from_hex_string(cls, hex_byte: str) -> OverviewStatus:
        """Create an instance from the hex string (e.g. ``'F1'``)"""
        return cls(*int_to_bits(int(hex_byte, base=16), n_bits=8))


class ErrorStatus(IntEnum):
    NoError = 0x00
    MotorCommunicationDisrupted = 0x01
    PlateNotMountedOnShovel = 0x02
    PlateNotDroppedFromShovel = 0x03
    ShovelNotExtended = 0x04
    ProcedureTimeout = 0x05
    TransferDoorNotOpened = 0x06
    TransferDoorNotClosed = 0x07
    ShovelNotRetracted = 0x08
    StepMotorTemperatureTooHigh = 0x0A
    OtherStepMotorError = 0x0B
    TransferStationNotRotated = 0x0C
    HeatingOrCo2CommunicationDisrupted = 0x0D
    ShakerCommunicationDisrupted = 0x0E
    ShakerConfigurationOutOfOrder = 0x0F
    ShakerNotStarted = 0x10
    ShakerClampNotOpen = 0x13
    ShakerClampNotClosed = 0x14
    Critical = 0xFF


class WarningStatus(IntEnum):
    NoWarning = 0x00
    MotorCommunicationDisrupted = 0x01
    PlateNotMountedOnShovel = 0x02
    PlateNotDroppedFromShovel = 0x03
    ShovelNotExtended = 0x04
    ProcedureTimeout = 0x05
    TransferDoorNotOpened = 0x06
    TransferDoorNotClosed = 0x07
    ShovelNotRetracted = 0x08
    InitialisingDueToOpenedDeviceDoor = 0x09
    TransferStationNotRotated = 0x0C


class ActionType(IntEnum):
    MoveHeightBelowSlot = 0x01
    CheckHeightBelowSlot = 0x02
    MoveHeightAboveSlot = 0x03
    CheckHeightAboveSlot = 0x04
    RotateToSlot = 0x05
    CheckRotation = 0x06
    ExtendShovel = 0x07
    CheckExtendedShovel = 0x08
    CheckShovelExtensionSensor = 0x09
    RetractShovel = 0x0A
    CheckRetractedShovel = 0x0B
    CloseTransferDoor = 0x0C
    CheckTransferDoorClosed = 0x0D
    OpenTransferDoor = 0x0E
    CheckTransferDoorOpened = 0x0F
    MoveSwapStationToPos1 = 0x10
    CheckSwapStationAtPos1 = 0x11
    MoveSwapStationToPos2 = 0x12
    CheckSwapStationAtPos2 = 0x13
    CheckPlateOnShovel = 0x14
    CheckPlateOnTransferStation = 0x15
    MoveToBarcodeReader = 0x16
    CheckHandlerAtBarcodeReader = 0x17
    ReadBarcode = 0x18


class ActionTarget(IntEnum):
    InitPosition = 1
    WaitPosition = 2
    Stacker = 3
    TransferStation = 4


class ActionStatus(NamedTuple):
    type: ActionType
    target: ActionTarget

    @classmethod
    def from_hex_string(cls, hex_byte: str) -> ActionStatus:
        """Create an instance from the hex string (e.g. ``'F1'``)"""
        num = int(hex_byte, base=16)
        type_ = enum_to_dict(ActionType)[(num & 0b11100000) >> 5]
        target = enum_to_dict(ActionTarget)[num & 0b00011111]
        return ActionStatus(type_, target)


class SwapStationStatus(NamedTuple):
    position1_at_door: bool
    occupied_at_door: bool
    occupied_at_user: bool

    @classmethod
    def from_response_string(cls, response: str) -> SwapStationStatus:
        """Create an instance from the response string (e.g. ``'111'``)"""
        return SwapStationStatus(
            position1_at_door=response[0] == "1",
            occupied_at_door=response[1] == "1",
            occupied_at_user=response[2] == "1",
        )



from contextlib import contextmanager
from enum import IntEnum
from threading import Lock
from typing import Dict, Iterator, Tuple, Type, TypeVar

GenericIntEnum = TypeVar("GenericIntEnum", bound=IntEnum)


@contextmanager
def lock_threading_lock(lock: Lock, *, timeout: float) -> Iterator[None]:
    """Context manager for using a `threading.lock` with a timeout"""
    if not timeout > 0:
        raise ValueError("Timeout must be positive")

    if not lock.acquire(timeout=timeout):
        raise TimeoutError(f"Could not acquire lock after {timeout} seconds")

    try:
        yield
    finally:
        lock.release()


def enum_to_dict(enum: Type[GenericIntEnum]) -> Dict[int, GenericIntEnum]:
    """Converts an enum class to a Dict[value, name]"""
    ret = {}
    for attr in dir(enum):
        item = getattr(enum, attr)
        if not isinstance(item, enum):
            continue
        ret[item.value] = item
    return ret


def int_to_bits(num: int, n_bits: int) -> Tuple[bool, ...]:
    """
    Convert an integer to its bit representation.

    Examples
    --------

    >>> int_to_bits(3, 2)
    (True, True)
    >>> int_to_bits(0xF1, 8)
    (True, True, True, True, False, False, False, True)
    """
    max_num_representable_by_n_bits = 2**n_bits - 1
    if num > max_num_representable_by_n_bits:
        raise ValueError(f"{n_bits} can only represent numbers <= {max_num_representable_by_n_bits}, got {num}")
    return tuple(bool(num & (2 ** (n_bits - i - 1))) for i in range(n_bits))



from cytomat.serial_port import SerialPort
from cytomat.status import OverviewStatus


class BarcodeScanner:
    __serial_port: SerialPort

    def __init__(self, serial_port: SerialPort) -> None:
        self.__serial_port = serial_port

    def read_all_barcodes(self) -> OverviewStatus:
        """
        Real all barcodes

        Returns
        -------
        The overview status after command submission
        """
        return self.__serial_port.issue_action_command("mv:sc")

    def read_barcodes_in_range(self, start_slot: int, end_slot: int) -> OverviewStatus:
        """
        Read all barcodes in the given range

        Parameters
        ----------
        start_slot
            The start slot
        end_slot
            The end slot

        Returns
        -------
        The overview status after command submission
        """
        return self.__serial_port.issue_action_command(f"mv:sc {start_slot:03} {end_slot:03}")

    def get_barcode_20char(self, slot: int) -> str:
        """
        Get the barcode at the given slot (for short barcodes)

        Parameters
        ----------
        slot
            The slot number

        Returns
        -------
        The barcode
        """
        return self.__serial_port.issue_status_command(f"ch:sc {slot:03}")

    def get_barcode_30char(self, slot: int) -> str:
        """
        Get the barcode at the given slot (for long barcodes)

        Parameters
        ----------
        slot
            The slot number

        Returns
        -------
        The barcode
        """
        return self.__serial_port.issue_status_command(f"ch:sd {slot:03}")

    def abort_barcode_reading(self) -> OverviewStatus:
        """
        Abort barcode reading

        Returns
        -------
        The overview status after command submission
        """
        return self.__serial_port.issue_action_command("rs:sc")

    @property
    def last_barcode_20char(self) -> str:
        """
        The last read barcode (for short barcodes)

        Returns
        -------
        The last barcode
        """
        return self.__serial_port.issue_status_command("ch:bc")

    @property
    def last_barcode_30char(self) -> str:
        """
        The last read barcode (for short barcodes)

        Returns
        -------
        The last barcode
        """
        return self.__serial_port.issue_status_command("ch:bd")

    def read_barcode_at_slot(self, slot: int) -> OverviewStatus:
        """
        Read the barcode at the given slot

        Parameters
        ----------
        slot
            The slot number

        Returns
        -------
        The overview status after command submission
        """
        return self.__serial_port.issue_action_command(f"ll:hb {slot:03}")

    def read_20char_barcode_at_current_position(self) -> OverviewStatus:
        """
        Read the barcode at the current position (for short barcodes)

        Returns
        -------
        The overview status after command submission
        """
        return self.__serial_port.issue_action_command("ll:bc")

    def read_30char_barcode_at_current_position(self) -> OverviewStatus:
        """
        Read the barcode at the current position (for long barcodes)

        Returns
        -------
        The overview status after command submission
        """
        return self.__serial_port.issue_action_command("ll:bd")
